import contextlib
import doctest
import importlib
import re
import shlex
import subprocess
import sys
from pathlib import Path
from typing import Iterator, TypedDict
from unittest.mock import patch

import freezegun
import lxml.doctestcompare
import lxml.etree
import lxml.objectify
import pkg_resources
import pytest
import yaml

import lowatt_enedis.services


def get_cases(_cache: dict[None, dict[str, str]] = {}) -> dict[str, str]:
    with contextlib.suppress(KeyError):
        return _cache[None]
    cases: dict[str, str] = {}
    with (Path(__file__).parent.parent / "doc" / "homologation.md").open() as f:
        for line in f:
            match = re.search(r"\|(.*)\|.*`(lowatt-enedis .*)`.*$", line)
            if match:
                case, command = match.groups()
                case = case.strip()
                assert case not in cases
                cases[case] = command
    assert len(cases) == 72
    _cache[None] = cases
    return cases


class ExpectedDict(TypedDict):
    url: str
    data: str


EXPECTED_FILENAME = Path(__file__).parent / "data" / "requests.yaml"


def get_expected(
    _cache: dict[None, dict[str, ExpectedDict]] = {}
) -> dict[str, ExpectedDict]:
    with contextlib.suppress(KeyError):
        return _cache[None]
    with EXPECTED_FILENAME.open() as f:
        result = _cache[None] = yaml.safe_load(f)
    return result  # type: ignore[no-any-return]


class Dumper(yaml.SafeDumper):
    pass


def multiline_dump(dumper: yaml.SafeDumper, data: str) -> yaml.ScalarNode:
    if len(data.splitlines()) > 1:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    return dumper.represent_scalar("tag:yaml.org,2002:str", data)


Dumper.add_representer(str, multiline_dump)


def set_expected(case: str, expected: ExpectedDict) -> None:
    try:
        with EXPECTED_FILENAME.open() as f:
            result = yaml.safe_load(f)
    except FileNotFoundError:
        result = {}

    result[case] = expected

    with EXPECTED_FILENAME.open("w") as f:
        yaml.dump(result, f, allow_unicode=True, sort_keys=True, Dumper=Dumper)


@contextlib.contextmanager
def override_sys_argv(argv: list[str]) -> Iterator[None]:
    old, sys.argv = sys.argv, argv
    try:
        yield
    finally:
        sys.argv = old


@pytest.fixture(scope="session")
def certs(tmp_path_factory: pytest.TempPathFactory) -> tuple[Path, Path]:
    certs = tmp_path_factory.mktemp("certs")
    certfile, keyfile = certs / "cert.pem", certs / "key.pem"
    subprocess.check_call(
        [
            "openssl",
            "req",
            "-x509",
            "-newkey",
            "rsa:2048",
            "-nodes",
            "-keyout",
            str(keyfile),
            "-out",
            str(certfile),
            "-batch",
        ],
    )
    return certfile, keyfile


def assert_xml_equal(actual: str, expected: str) -> None:
    checker = lxml.doctestcompare.LXMLOutputChecker()
    if not checker.check_output(actual, expected, 0):
        message = checker.output_difference(doctest.Example("", expected), actual, 0)
        raise AssertionError(message)


@pytest.mark.parametrize("case", list(get_cases()))
def test_requests(
    case: str, tmp_path: Path, certs: tuple[Path, Path], regen_test_data: bool
) -> None:
    # fix some defaults set at import time
    with freezegun.freeze_time("2022-02-22"), patch.dict(
        "os.environ", {"ENEDIS_CONTRAT": "1234"}
    ):
        importlib.reload(lowatt_enedis.services)
    entrypoint = pkg_resources.get_entry_info(
        "lowatt_enedis",
        "console_scripts",
        "lowatt-enedis",
    )
    assert entrypoint
    func = entrypoint.load()
    command = shlex.split(get_cases()[case])

    with override_sys_argv(command), patch.dict(
        "os.environ",
        {
            "ENEDIS_LOGIN": "test@example.com",
            "ENEDIS_CERT_FILE": str(certs[0]),
            "ENEDIS_KEY_FILE": str(certs[1]),
            "ENEDIS_HOMOLOGATION": "true",
        },
    ), patch("lowatt_enedis.certauth._HTTPSClientAuthHandler.https_open") as client:

        class FakeResponse:
            code = 500
            msg = b""

            def info(self) -> dict[str, str]:
                return {}

            def read(self) -> bytes:
                return (
                    b'<?xml version="1.0" ?>'
                    b'<soapenv:Envelope xmlns:soapenv="http://schemas.xmlsoap.org/soap/envelope/">'
                    b'<soap:Body xmlns:soap="http://schemas.xmlsoap.org/soap/envelope/">'
                    b'<ns3:Fault xmlns:ns3="http://schemas.xmlsoap.org/soap/envelope/">'
                    b"</ns3:Fault>"
                    b"</soap:Body>"
                    b"</soapenv:Envelope>"
                )

            def close(self) -> None:
                pass

        resp = FakeResponse()
        client.return_value = resp
        with pytest.raises(SystemExit) as cm:
            func()
        assert cm.value.code == 1
    if sys.version_info[:2] <= (3, 7):
        req = list(client.mock_calls[0])[1][0]
    else:
        req = client.mock_calls[0].args[0]
    data = lxml.etree.tostring(
        lxml.objectify.fromstring(req.data), encoding="utf8", pretty_print=True
    ).decode()
    actual: ExpectedDict = {"url": req.full_url, "data": data}
    if regen_test_data:
        set_expected(case, actual)
    else:
        expected = get_expected()[case]
        assert actual["url"] == expected["url"]
        assert_xml_equal(actual["data"], expected["data"])



# Copyright (c) 2021 by Lowatt - info@lowatt.fr
#
# This program is part of lowatt_enedis
#
# lowatt_enedis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lowatt_enedis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lowatt_enedis.  If not, see <https://www.gnu.org/licenses/>.
"""
`lowatt_enedis.services`
------------------------

SGE web-service mapping to plug them into the CLI.
"""

import argparse
import warnings
from datetime import date, datetime, timedelta
from typing import Any, Iterator, Literal

import suds.sudsobject
from dateutil import tz
from suds.client import Client

from . import (
    arg_from_env,
    create_from_options,
    dict_from_dicts,
    get_option,
    register,
    ws,
)

UTC = tz.tzutc()
ACCORD_CLIENT_OPTIONS = {
    # XXX exclusive option groups
    "--denomination": {
        "help": "Dénomination sociale de la personne morale ayant donné son accord",
    },
    "--nom": {
        "help": "Nom de la personne physique ou nom commerciale de la "
        "personne morale ayant donné son accord",
    },
    "--prenom": {
        "help": "Prénom de la personne physique",
    },
    "--civilite": {
        "choices": ["M", "Mme", "Mlle"],
        "help": "Civilité de la personne physique",
    },
    "--no-autorisation": {
        "action": "store_true",
        "help": "demander sans l'autorisation du client (pour l'homologation seulement)",
    },
}
ACCORD_CLIENT_EXTENDED_OPTIONS = {
    "--naf": {
        "help": "Code NAF de l'activité de la personne morale",
    },
    "--secteur": {
        "help": "Code du secteur d'activité de la personne morale",
    },
    "--siret": {
        "help": "Numéro de SIRET de l'établissement principal de la personne morale",
    },
}

CONTRAT_OPTIONS = {
    "--contrat": dict_from_dicts(
        {
            "help": (
                "identifiant du contrat entre le demandeur et Enedis. "
                "Default to ENEDIS_CONTRAT environment variable."
            ),
        },
        arg_from_env("ENEDIS_CONTRAT"),
    ),
}


def _accord_client(
    client: Client,
    args: argparse.Namespace,
    extended: bool = False,
    xs_accord_type: str = "ns3:DeclarationAccordClientType",
    autorisation: bool = True,
) -> suds.sudsobject.Object:
    if get_option(args, "denomination"):
        ptype = "Morale"
        who_options = {
            "denomination": "denominationSociale",
        }
        unavailable_options = {"prenom", "civilite"}
        if extended:
            who_options.update(
                {
                    "nom": "nomCommercial",
                    "naf": "activiteCodeNaf",
                    "secteur": "secteurActivite",
                    "siret": "etablissementPrincipalNumSiret",
                },
            )
    elif get_option(args, "nom"):
        ptype = "Physique"
        who_options = {
            "nom": "nom",
            "prenom": "prenom",
            "civilite": "civilite",
        }
        unavailable_options = {"naf", "secteur", "siret"}
    elif extended:  # person info is not mandatory
        ptype = None
    else:
        raise ValueError(
            "You should specify at least one of 'denomination' or 'nom' option",
        )

    accord = client.factory.create(xs_accord_type)

    if hasattr(accord, "accordClient"):
        accord.accordClient = _boolean(autorisation)
    elif hasattr(accord, "accord"):
        # tag name used by CommandeCollectePublicationMesures
        accord.accord = _boolean(autorisation)

    if ptype is not None:
        for option in unavailable_options:
            if getattr(args, option, None):
                raise ValueError(
                    f"L'option {option} ne peut être spécifiée que pour une personne "
                    f"{ptype.lower()}.",
                )

        who = create_from_options(client, args, f"Personne{ptype}Type", who_options)
        setattr(accord, f"personne{ptype}", who)

    return accord


MESURES_OPTIONS = {
    "--from": {
        "default": (date.today() - timedelta(days=1)).isoformat(),
        "help": "date de début souhaitée (incluse)",
    },
    "--to": {
        "default": date.today().isoformat(),
        "help": "date de fin souhaitée (excluse)",
    },
    "--corrigee": {
        "action": "store_true",
        "help": "obtenir les mesures corrigées (C1-C4 uniquement)",
    },
}

MESURES_OPTIONS_MAP = {
    "from": "dateDebut",
    "to": "dateFin",
    "corrigee": "mesuresCorrigees",
}


def _pas(
    client: Client, args: argparse.Namespace, demande: suds.sudsobject.Object
) -> None:
    pas = get_option(args, "pas")
    if pas:
        demande.pasCdc = client.factory.create("ns1:DureeType")
        demande.pasCdc.unite = "min"
        demande.pasCdc.valeur = pas


def _check_mesure_options_consistency(args: argparse.Namespace) -> None:
    if get_option(args, "pas") and get_option(args, "type") != "CDC":
        raise ValueError(
            "L'option 'pas' ne peut être  spécifiée que pour le type de "
            "mesure 'courbe de charge' (CDC).",
        )


@register(
    "search",
    "RecherchePoint-v2.0",
    {
        "--siret": {
            "help": "n° de siret",
        },
        "--compteur": {
            "help": "matricule ou numéro de série du compteur",
        },
        "--tension": {
            "choices": ["BTINF", "BTSUP", "HTA", "HTB"],
            "help": "domaine de tension d'alimentation",
        },
        "--nom": {
            "help": "nom ou dénomination sociale du client final",
        },
        "--categorie": {
            "choices": ["PRO", "RES"],
            "help": "catégorie de client final",
        },
        "--hp": {
            "action": "store_true",
            "help": "recherche hors-périmetre",
        },
        "--etage": {
            "help": "escalier et étage appartement",
        },
        "--batiment": {
            "help": "batiment",
        },
        "--voie": {
            "help": "numéro et nom de la voie",
        },
        "--lieuDit": {
            "help": "lieu-dit",
        },
        "--cp": {
            "help": "code postal",
        },
        "--insee": {
            "help": "code INSEE de la commune",
        },
    },
)
@ws("RecherchePoint-v2.0")
def search_point(client: Client, args: argparse.Namespace) -> suds.sudsobject.Object:
    criteria = create_from_options(
        client,
        args,
        "CriteresType",
        {
            "siret": "numSiret",
            "compteur": "matriculeOuNumeroSerie",
            "tension": "domaineTensionAlimentationCode",
            "nom": "nomClientFinalOuDenominationSociale",
            "categorie": "categorieClientFinalCode",
            "hp": "rechercheHorsPerimetre",
        },
    )
    assert criteria is not None
    address = create_from_options(
        client,
        args,
        "AdresseInstallationType",
        {
            "etage": "escalierEtEtageEtAppartement",
            "batiment": "batiment",
            "voie": "numeroEtNomVoie",
            "lieuDit": "lieuDit",
            "cp": "codePostal",
            "insee": "codeInseeCommune",
        },
    )
    if address is not None:
        criteria.adresseInstallation = address
    return client.service.rechercherPoint(criteria, get_option(args, "login"))


@register(
    "technical",
    "ConsultationDonneesTechniquesContractuelles-v1.0",
    {
        "prm": {
            "help": "identifiant PRM du point",
        },
        "--autorisation": {
            "help": "indique l'autorisation du client",
            "action": "store_true",
            "default": False,
        },
    },
)
@ws("ConsultationDonneesTechniquesContractuelles-v1.0")
def point_technical_data(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    return client.service.consulterDonneesTechniquesContractuelles(
        get_option(args, "prm"),
        get_option(args, "login"),
        _boolean(get_option(args, "autorisation")),
    )


@register(
    "measures",
    "ConsultationMesures-v1.1",
    dict_from_dicts(
        {
            "prm": {
                "help": "identifiant PRM du point",
            },
            "--autorisation": {
                "help": "indique l'autorisation du client",
                "action": "store_true",
                "default": False,
            },
        },
        CONTRAT_OPTIONS,
    ),
)
@ws("ConsultationMesures-v1.1")
def point_measures(client: Client, args: argparse.Namespace) -> suds.sudsobject.Object:
    return client.service.consulterMesures(
        get_option(args, "prm"),
        get_option(args, "login"),
        get_option(args, "contrat"),
        _boolean(get_option(args, "autorisation")),
    )


def parse_date(value: str) -> date:
    return date(*(int(part) for part in value.split("-")))


def measures_resp2py(resp: suds.sudsobject.Object) -> Iterator[dict[str, Any]]:
    """Return a list of dictionaries given a response from
    ConsultationMesures web-service. Each dict contains the following keys:
    - `grille`: "frn" or "turpe",
    - `grandeurPhysique`: "EA" for active energy,
    - `classeTemporelle`: eg. "BASE", "HP", "HC",
    - `calendrier`: calendar id,
    - `unit`: "kWh",
    - `mesures`: a list of point, each one being a 5-uple with:

      - begin date,
      - end date,
      - value,
      - nature,
      - trigger,
      - status.
    """
    for attr in ("seriesMesuresDateesGrilleTurpe", "seriesMesuresDateesGrilleFrn"):
        try:
            series = getattr(resp, attr).serie
        except AttributeError:
            continue
        grille = attr.split("Grille")[1].lower()
        for serie in series:
            mesures = []
            for point in serie.mesuresDatees.mesure:
                mesures.append(
                    (
                        parse_date(point.dateDebut),
                        parse_date(point.dateFin),
                        int(point.valeur),
                        point.nature._code,
                        point.declencheur._code,
                        point.statut._code,
                    ),
                )
            try:
                calendrier = serie.calendrier._code
            except AttributeError:
                calendrier = None
            yield {
                "grille": grille,
                "grandeurPhysique": serie.grandeurPhysique._code,
                "classeTemporelle": serie.classeTemporelle._code,
                "calendrier": calendrier,
                "unit": serie.unite,
                "mesures": mesures,
            }


@register(
    "details",
    "ConsultationMesuresDetaillees-v2.0",
    dict_from_dicts(
        {
            "prm": {
                "help": "identifiant PRM du point",
            },
            "type": {
                "choices": ["ENERGIE", "PMAX", "COURBE"],
                "help": "type de mesure demandé : ENERGIE pour les consommations "
                "globales quotidiennes, PMAX pour les puissances maximales "
                "quotidiennes, COURBE pour la courbe de charge.",
            },
            "--courbe-type": {
                # - PA pour récupérer les courbes de puissance active (seule
                #   courbe disponible pour les segments C5 et P4)
                # - PRI pour récupérer les courbes de puissance réactive
                #   inductive
                # - PRC pour récupérer les courbes de puissance réactive
                #   capacitive
                # - E pour récupérer les courbes de tension
                # - TOUT pour récupérer les courbes disponibles
                "choices": ["PA", "PRI", "PRC", "E", "TOUT"],
                "default": "PA",
                "help": "type de courbe demandé le cas échéant (Puissance Active, "
                "Puissance Réactive Inductive, Puissance Réactive Capacitive, "
                "tension (E), tout).",
            },
            "--injection": {
                "action": "store_true",
                "help": "demander les données en injection (soutirage par défaut).",
            },
            "--no-autorisation": {
                "action": "store_true",
                "help": "demander sans l'autorisation du client (pour l'homologation seulement)",
            },
        },
        MESURES_OPTIONS,
    ),
)
@ws("ConsultationMesuresDetaillees-v2.0")
def point_detailed_measures(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    warnings.warn(
        "ConsultationMesuresDetaillees-v2.0 is deprecated and pending for removal",
        DeprecationWarning,
        stacklevel=2,
    )
    demande = create_from_options(
        client,
        args,
        "DemandeType",
        dict_from_dicts(
            {
                "login": "initiateurLogin",
                "prm": "pointId",
                "type": "mesuresTypeCode",
            },
            MESURES_OPTIONS_MAP,
        ),
    )
    assert demande is not None
    # demande.mesuresPas uniquement pour PMAX sur C5N, valeur P1D ou P1M
    if get_option(args, "type") == "PMAX":
        demande.grandeurPhysique = "PMA"
        demande.mesuresPas = "P1D"  # 'P1M'
    else:
        del demande.mesuresPas
        if get_option(args, "type") == "ENERGIE":
            demande.grandeurPhysique = "EA"
        elif get_option(args, "type") == "COURBE":
            courbe_type = get_option(args, "courbe_type")
            if not courbe_type:
                raise ValueError(
                    "l'option courbe-type doit être spécifié pour le type de mesure COURBE",
                )
            demande.grandeurPhysique = courbe_type
    injection = get_option(args, "injection")
    demande.soutirage = _boolean(not injection)
    demande.injection = _boolean(injection)
    demande.accordClient = _boolean(not get_option(args, "no_autorisation"))

    return client.service.consulterMesuresDetaillees(demande)


@register(
    "detailsV3",
    "ConsultationMesuresDetaillees-v3.0",
    dict_from_dicts(
        {
            "prm": {
                "help": "identifiant PRM du point",
            },
            "type": {
                "choices": ["ENERGIE", "PMAX", "COURBE", "INDEX"],
                "help": "type de mesure demandé : ENERGIE pour les consommations "
                "globales quotidiennes, PMAX pour les puissances maximales "
                "quotidiennes, COURBE pour la courbe de charge.",
            },
            "--courbe-type": {
                # - PA pour récupérer les courbes de puissance active (seule
                #   courbe disponible pour les segments C5 et P4)
                # - PRI pour récupérer les courbes de puissance réactive
                #   inductive
                # - PRC pour récupérer les courbes de puissance réactive
                #   capacitive
                # - E pour récupérer les courbes de tension
                # - TOUT pour récupérer les courbes disponibles
                "choices": ["PA", "PRI", "PRC", "E", "TOUT"],
                "default": "PA",
                "help": "type de courbe demandé le cas échéant (Puissance Active, "
                "Puissance Réactive Inductive, Puissance Réactive Capacitive, "
                "tension (E), tout).",
            },
            "--injection": {
                "action": "store_true",
                "help": "demander les données en injection (soutirage par défaut).",
            },
            "--cadre": {
                "choices": ["ACCORD_CLIENT", "SERVICE_ACCES", "EST_TITULAIRE"],
                "default": "ACCORD_CLIENT",
                "help": "Cadre d'accès à la demande, ACCORD_CLIENT par défaut.",
            },
        },
        MESURES_OPTIONS,
    ),
)
@ws("ConsultationMesuresDetaillees-v3.0")
def point_detailed_measuresV3(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    demande = create_from_options(
        client,
        args,
        "Demande",
        dict_from_dicts(
            {
                "login": "initiateurLogin",
                "prm": "pointId",
                "type": "mesuresTypeCode",
            },
            MESURES_OPTIONS_MAP,
        ),
    )
    assert demande is not None
    # demande.mesuresPas uniquement pour PMAX sur C5N, valeur P1D ou P1M
    if get_option(args, "type") == "PMAX":
        demande.grandeurPhysique = "PMA"
        demande.mesuresPas = "P1D"  # 'P1M'
    else:
        del demande.mesuresPas
        if get_option(args, "type") in ("ENERGIE", "INDEX"):
            demande.grandeurPhysique = "EA"
        elif get_option(args, "type") == "COURBE":
            courbe_type = get_option(args, "courbe_type")
            if not courbe_type:
                raise ValueError(
                    "l'option courbe-type doit être spécifié pour le type de mesure COURBE",
                )
            demande.grandeurPhysique = courbe_type
    injection = get_option(args, "injection")
    demande.sens = "INJECTION" if injection else "SOUTIRAGE"
    demande.cadreAcces = get_option(args, "cadre")
    return client.service.consulterMesuresDetailleesV3(demande)


def detailed_measures_resp2py(
    resp: suds.sudsobject.Object,
) -> Iterator[tuple[datetime, float]]:
    """Return an iterator on (UTC tz naive datetime, value) given a response from
    ConsultationMesuresDetaillees web-service.
    """
    # start = resp.periode.dateDebut
    # end = resp.periode.dateFin
    assert len(resp.grandeur) == 1
    for point in resp.grandeur[0].mesure:
        yield (point.d.astimezone(UTC), point.v)


def _donnees_generales(
    client: Client,
    args: argparse.Namespace,
    code: str,
    tag_name: str,
) -> suds.sudsobject.Object:
    obj = create_from_options(
        client,
        args,
        tag_name,
        {
            # XXX missing: refFrn, refFrnRegroupement, affaireLieeId
            "contrat": "contratId",
            "prm": "pointId",
            "login": "initiateurLogin",
        },
    )
    assert obj is not None
    obj.objetCode = code
    return obj


DONNEES_GENERALES_OPTIONS = dict_from_dicts(
    {
        "prm": {
            "help": "identifiant PRM du point",
        },
    },
    CONTRAT_OPTIONS,
)


@register(
    "cmdHisto",
    "CommandeTransmissionHistoriqueMesures-v1.0",
    dict_from_dicts(
        DONNEES_GENERALES_OPTIONS,
        {
            "type": {
                "choices": ["IDX", "CDC"],
                "help": "type de mesure demandé : IDX pour les index, CDC pour la "
                "courbe de charge.",
            },
            "--pas": {
                "choices": ["10", "30"],
                "help": "pas souhaité dans le cas des courbes de charges ; "
                "10 pour C1-C4 / 30 pour C5",
            },
        },
        ACCORD_CLIENT_OPTIONS,
        ACCORD_CLIENT_EXTENDED_OPTIONS,
        MESURES_OPTIONS,
        {
            # override --from's default to 2 years
            "--from": {
                "default": (date.today() - timedelta(days=365 * 2)).isoformat(),
                "help": "date de début souhaitée",
            },
        },
    ),
)
@ws("CommandeTransmissionHistoriqueMesures-v1.0")
def point_cmd_histo(client: Client, args: argparse.Namespace) -> suds.sudsobject.Object:
    warnings.warn(
        "CommandeTransmissionHistoriqueMesures-v1.0 is deprecated and pending for removal",
        DeprecationWarning,
        stacklevel=2,
    )
    _check_mesure_options_consistency(args)

    header = client.options.soapheaders
    header.infoFonctionnelles.pointId = get_option(args, "prm")
    del header.infoFonctionnelles.affaireId

    demande = client.factory.create("ns0:HistoriqueMesuresDemandeType")

    demande.donneesGenerales = _donnees_generales(
        client,
        args,
        "HDM",
        "DemandeDonneesGeneralesType",
    )

    demande.historiqueMesures = create_from_options(
        client,
        args,
        "DemandeHistoriqueMesuresType",
        dict_from_dicts(
            MESURES_OPTIONS_MAP,
            {
                "type": "mesureType",
                "corrigee": "mesureCorrigee",  # override mesuresCorrigees
            },
        ),
    )
    assert demande.historiqueMesures is not None
    _pas(client, args, demande.historiqueMesures)
    demande.historiqueMesures.declarationAccordClient = _accord_client(
        client,
        args,
        True,
        "ns0:DeclarationConsentementType",
    )

    return client.service.commanderTransmissionHistoriqueMesures(demande)


@register(
    "cmdAcces",
    "CommanderAccesDonneesMesures-V1.0",
    dict_from_dicts(
        DONNEES_GENERALES_OPTIONS,
        {
            "--from": {
                "default": (date.today()).isoformat(),
                "help": "date de début souhaitée (incluse)",
            },
            "--to": {
                "default": (date.today() + timedelta(days=365 * 3)).isoformat(),
                "help": "date de fin souhaitée (excluse)",
            },
            "type": {
                "choices": ["ENERGIE", "PMAX", "COURBE", "INDEX"],
                "help": "type de mesure demandé : ENERGIE pour les consommations "
                "globales quotidiennes, PMAX pour les puissances maximales "
                "quotidiennes, COURBE pour la courbe de charge.",
            },
            "--injection": {
                "action": "store_true",
                "help": "demander les données en injection (soutirage par défaut).",
            },
        },
        ACCORD_CLIENT_OPTIONS,
    ),
)
@ws("CommanderAccesDonneesMesures-V1.0")
def point_cmd_acces(client: Client, args: argparse.Namespace) -> suds.sudsobject.Object:
    demande = client.factory.create("ns1:DemandeType")

    demande.donneesGenerales = create_from_options(
        client,
        args,
        "DonneesGeneralesType",
        {
            # XXX missing: refExterne
            "prm": "pointId",
            "login": "initiateurLogin",
        },
    )
    assert demande.donneesGenerales is not None

    demande.donneesGenerales.objetCode = "AME"
    demande.donneesGenerales.contrat = create_from_options(
        client,
        args,
        "ContratType",
        {
            # XXX missing: acteurMarcheCode, contratType
            "contrat": "contratId",
        },
    )

    demande.accesDonnees = create_from_options(
        client,
        args,
        "AccesDonneesType",
        {
            "from": "dateDebut",
            "to": "dateFin",
            "type": "typeDonnees",
        },
    )
    assert demande.accesDonnees is not None

    # COURBE and INDEX are correct values according to the documentation, but
    # the error "SGT4Q1: Le type de mesure renseigné n'est pas conforme." is returned.
    # CDC and IDX do not return this error.
    if demande.accesDonnees.typeDonnees == "COURBE":
        demande.accesDonnees.typeDonnees = "CDC"
    elif demande.accesDonnees.typeDonnees == "INDEX":
        demande.accesDonnees.typeDonnees = "IDX"

    injection = get_option(args, "injection")
    demande.accesDonnees.soutirage = _boolean(not injection)
    demande.accesDonnees.injection = _boolean(injection)

    demande.accesDonnees.declarationAccordClient = _accord_client(
        client,
        args,
        xs_accord_type="ns1:DeclarationAccordClientType",
        autorisation=not get_option(args, "no_autorisation"),
    )

    return client.service.commanderAccesDonneesMesures(demande)


@register(
    "cmdInfraJ",
    "CommandeTransmissionDonneesInfraJ-v1.0",
    dict_from_dicts(
        DONNEES_GENERALES_OPTIONS,
        {
            "--injection": {
                "action": "store_true",
                "help": "demander les données en injection (courbe de charge, "
                "tension et index).",
            },
            "--soutirage": {
                "action": "store_true",
                "help": "demander les données en soutirage (courbe de charge, "
                "tension et index et paramètres de tarification dynamique).",
            },
            "--cdc": {
                "action": "store_true",
                "help": "demander les données les courbes de charge et de tension.",
            },
            "--idx": {
                "action": "store_true",
                "help": "demander les données les données d'index.",
            },
            "--ptd": {
                "action": "store_true",
                "help": "demander les paramètres de tarification dynamique.",
            },
        },
        ACCORD_CLIENT_OPTIONS,
    ),
)
@ws("CommandeTransmissionDonneesInfraJ-v1.0")
def point_cmd_infra_j(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    demande = client.factory.create("ns1:DemandeType")

    demande.donneesGenerales = _donnees_generales(
        client,
        args,
        "AME",
        "DonneesGeneralesType",
    )

    demande.accesDonnees = create_from_options(
        client,
        args,
        "DemandeAccesDonneesType",
        {
            "injection": "injection",
            "soutirage": "soutirage",
            "cdc": "cdc",
            "idx": "idx",
            "ptd": "ptd",
        },
    )
    assert demande.accesDonnees is not None
    demande.accesDonnees.declarationAccordClient = accord = _accord_client(
        client,
        args,
        xs_accord_type="ns1:DeclarationAccordClientType",
        autorisation=not get_option(args, "no_autorisation"),
    )
    accord.injection = _boolean(get_option(args, "injection"))
    accord.soutirage = _boolean(get_option(args, "soutirage"))
    return client.service.commanderTransmissionDonneesInfraJ(demande)


@register(
    "subscribe",
    "CommandeCollectePublicationMesures-v3.0",
    dict_from_dicts(
        DONNEES_GENERALES_OPTIONS,
        {
            "--cdc-enable": {
                "action": "store_true",
                "help": "demander la collecte de la courbe de charge.",
            },
            "--cdc": {
                "action": "store_true",
                "help": "demander la transmission données les courbes de charge.",
            },
            "--idx": {
                "action": "store_true",
                "help": "demander la transmission des données d'index index et autres "
                "données du compteur.",
            },
            "--period": {
                "choices": ["daily", "weekly", "monthly"],
                "default": "daily",
                "help": "périodicité de réception pour la courbe de charge "
                "(quotidienne par défaut).",
            },
            "--injection": {
                "action": "store_true",
                "help": "demander les données en injection (soutirage par défaut).",
            },
            "--linky": {
                "action": "store_true",
                "help": "la demande concerne le segment C5 ou P4"
                "(segment C1-C4 ou P1-P3 par défaut).",
            },
        },
        ACCORD_CLIENT_OPTIONS,
        MESURES_OPTIONS,
        {
            # override --from --to
            "--from": {
                "default": date.today().isoformat(),
                "help": "date de début, postérieure ou égale à la date du jour",
            },
            "--to": {
                "default": (date.today() + timedelta(days=365)).isoformat(),
                "help": "date de fin souhaitée, "
                "supérieure à la date de fin précédente lors d’un renouvellement, "
                "un an max pour les compteurs linky",
            },
        },
    ),
)
@ws("CommandeCollectePublicationMesures-v3.0")
def point_cmd_publication(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    demande = client.factory.create("ns1:DemandeType")

    demande.donneesGenerales = _donnees_generales(
        client,
        args,
        "AME",
        "DonneesGeneralesType",
    )

    demande.accesMesures = acces = create_from_options(
        client,
        args,
        "DemandeAccesMesures",
        dict_from_dicts(MESURES_OPTIONS_MAP),
    )
    assert demande.accesMesures is not None
    assert acces is not None

    supported_actions = ["cdc_enable", "cdc", "idx"]
    actions = [a for a in supported_actions if get_option(args, a)]

    if not actions or len(actions) > 1:
        raise ValueError(f"Une action doit être selectionnée parmi {supported_actions}")

    action = actions[0]

    if action == "idx":
        code = "IDX"
        transmission = True
        pas = "P1D"
        period = "P1D"
    else:
        code = "CDC"
        transmission = action == "cdc"
        if get_option(args, "linky"):
            pas = "PT30M"
        else:
            pas = "PT10M"
        period = get_option(args, "period")
        if period == "monthly":
            period = "P1M"
        elif period == "weekly":
            period = "P7D"
        else:
            period = "P1D"

    injection = get_option(args, "injection")

    acces.mesuresTypeCode = code
    acces.transmissionRecurrente = _boolean(transmission)
    acces.mesuresPas = pas
    acces.periodiciteTransmission = period
    acces.injection = _boolean(injection)
    acces.soutirage = _boolean(not injection)

    if injection and get_option(args, "linky"):
        del demande.periodiciteTransmission

    demande.accesMesures.declarationAccordClient = _accord_client(
        client,
        args,
        xs_accord_type="ns1:DeclarationAccordClientType",
        autorisation=not get_option(args, "no_autorisation"),
    )

    return client.service.commanderCollectePublicationMesures(demande)


@register(
    "subscriptions",
    "RechercherServicesSouscritsMesures-v1.0",
    DONNEES_GENERALES_OPTIONS,
)
@ws("RechercherServicesSouscritsMesures-v1.0")
def point_search_subscriptions(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    criteria = create_from_options(
        client,
        args,
        "CriteresType",
        {
            "contrat": "contratId",
            "prm": "pointId",
        },
    )

    return client.service.rechercherServicesSouscritsMesures(
        criteria,
        get_option(args, "login"),
    )


@register(
    "unsubscribe",
    "CommandeArretServiceSouscritMesures-v1.0",
    dict_from_dicts(
        DONNEES_GENERALES_OPTIONS,
        {
            "--id": {
                "help": "identifiant du service souscrit de mesures à arrêter",
            },
        },
    ),
)
@ws("CommandeArretServiceSouscritMesures-v1.0")
def point_unsubscribe(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    demande = client.factory.create("ns1:DemandeType")

    demande.donneesGenerales = _donnees_generales(
        client,
        args,
        "ASS",
        "DonneesGeneralesType",
    )

    demande.arretServiceSouscrit = create_from_options(
        client,
        args,
        "ArretServiceSouscritType",
        {"id": "serviceSouscritId"},
    )

    return client.service.commanderArretServiceSouscritMesures(demande)


M023_COMMON_OPTIONS = dict_from_dicts(
    {
        "prms": {
            "help": "identifiants PRM des point, séparés par des virgules",
        },
        "--sens": {
            "choices": ["SOUTIRAGE", "INJECTION"],
            "default": "SOUTIRAGE",
            "help": "Sens de l’énergie circulant vers le réseau d’Enedis, "
            "SOUTIRAGE par défaut.",
        },
        "--cadre": {
            "choices": ["ACCORD_CLIENT", "SERVICE_ACCES", "EST_TITULAIRE"],
            "default": "ACCORD_CLIENT",
            "help": "Cadre d'accès à la demande, ACCORD_CLIENT par défaut.",
        },
        "--format": {
            "choices": ["JSON", "CSV"],
            "default": "JSON",
            "help": "Format des fichiers de résultats de la demande attendu, "
            "JSON par défaut",
        },
    },
    CONTRAT_OPTIONS,
)
M023_LOGIN_OPTIONS_MAP = {
    "login": "initiateurLogin",
    "contrat": "contratId",
}
M023_COMMON_OPTIONS_MAP = {
    "sens": "sens",
    "cadre": "cadreAcces",
    "format": "format",
}


# CommandeHistoriqueDonneesMesuresFines
@register(
    "cmdHistoFine",
    "B2B_M023MFI",
    dict_from_dicts(
        M023_COMMON_OPTIONS,
        {
            "type": {
                "choices": ["ENERGIE", "PMAX", "COURBES", "INDEX"],
                "help": "type de mesure demandé : ENERGIE pour les énergies "
                "globales quotidiennes, PMAX » pour les puissances maximales "
                "quotidiennes ou mensuelles, COURBES pour une courbe de "
                "puissance ou de tension, INDEX pour les index quotidiens. ",
            },
        },
        MESURES_OPTIONS,
    ),
)
@ws("B2B_M023MFI")
def point_cmd_histo_fine(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    donneesGenerales = create_from_options(
        client, args, "donneesGenerales", M023_LOGIN_OPTIONS_MAP
    )
    assert donneesGenerales is not None

    demande = create_from_options(
        client,
        args,
        "demande",
        dict_from_dicts(
            {"type": "mesuresTypeCode"},
            M023_COMMON_OPTIONS_MAP,
            MESURES_OPTIONS_MAP,
        ),
    )
    assert demande is not None

    demande.pointIds.pointId = get_option(args, "prms").split(",")

    return client.service.commandeHistoriqueDonneesMesuresFines(
        donneesGenerales, demande
    )


# CommandeHistoriqueDonneesMesuresFacturantes
@register(
    "cmdHistoFact",
    "B2B_M023MFA",
    dict_from_dicts(
        M023_COMMON_OPTIONS,
        {
            # MESURES_OPTIONS without --corrigee
            "--from": MESURES_OPTIONS["--from"],
            "--to": MESURES_OPTIONS["--to"],
        },
    ),
)
@ws("B2B_M023MFA")
def point_cmd_histo_fact(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    donneesGenerales = create_from_options(
        client, args, "donneesGenerales", M023_LOGIN_OPTIONS_MAP
    )
    assert donneesGenerales is not None

    demande = create_from_options(
        client,
        args,
        "demande",
        dict_from_dicts(
            M023_COMMON_OPTIONS_MAP,
            {
                "from": MESURES_OPTIONS_MAP["from"],
                "to": MESURES_OPTIONS_MAP["to"],
            },
        ),
    )
    assert demande is not None

    demande.pointIds.pointId = get_option(args, "prms").split(",")

    return client.service.commandeHistoriqueDonneesMesuresFacturantes(
        donneesGenerales, demande
    )


# CommandeInformationsTechniquesEtContractuelles
@register(
    "cmdTechnical",
    "B2B_M023ITC",
    M023_COMMON_OPTIONS,
)
@ws("B2B_M023ITC")
def point_cmd_technical(
    client: Client, args: argparse.Namespace
) -> suds.sudsobject.Object:
    donneesGenerales = create_from_options(
        client, args, "donneesGenerales", M023_LOGIN_OPTIONS_MAP
    )
    assert donneesGenerales is not None

    demande = create_from_options(client, args, "demande", M023_COMMON_OPTIONS_MAP)
    assert demande is not None

    demande.pointIds.pointId = get_option(args, "prms").split(",")

    return client.service.commandeInformationsTechniquesEtContractuelles(
        donneesGenerales, demande
    )


def _boolean(b: Any) -> Literal["true", "false"]:
    return "true" if b else "false"



# Copyright (c) 2021 by Lowatt - info@lowatt.fr
#
# This program is part of lowatt_enedis
#
# lowatt_enedis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lowatt_enedis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lowatt_enedis.  If not, see <https://www.gnu.org/licenses/>.
"""
`lowatt_enedis`
---------------

Command line interface to enedis SGE web-services.
"""

import argparse
import datetime
import json
import logging
import os
import sys
import xml.dom.minidom
from functools import wraps
from pathlib import Path, PurePath
from typing import Any, Callable, Iterator, Optional, TypeVar, Union

import rich
import suds.sudsobject
from suds import WebFault
from suds.client import Client
from suds.plugin import DocumentPlugin

from .certauth import HTTPSClientCertTransport

RT = TypeVar("RT")

logging.basicConfig(level=logging.INFO)
# send raw enedis errors to stderr:
errHandler = logging.StreamHandler(sys.stderr)
logging.getLogger("suds.client").addHandler(errHandler)
logging.getLogger("suds.client").propagate = False
# logging.getLogger("suds.client").setLevel(logging.DEBUG)
# logging.getLogger("suds.resolver").setLevel(logging.DEBUG)
# logging.getLogger("suds.mx").setLevel(logging.DEBUG)
# logging.getLogger("suds.transport").setLevel(logging.DEBUG)


WSDL_DIR = PurePath(__file__).parent.joinpath("wsdl")
SERVICES = {x.stem: x.resolve().as_uri() for x in Path(WSDL_DIR).glob("**/*.wsdl")}
COMMAND_SERVICE: dict[str, tuple[str, dict[str, Any], Any]] = {}


def wsdl(service_name: str) -> str:
    """Return path to WSDL file for `service_name`."""
    try:
        return SERVICES[service_name]
    except KeyError:
        raise KeyError(
            "Unknown service name {!r}, available services are {}".format(
                service_name,
                ", ".join(sorted(SERVICES)),
            ),
        ) from None


def arg_from_env(key: str) -> dict[str, Any]:
    default = os.environ.get(key)
    return {"default": default, "required": not default}


def init_cli(subparsers: Any) -> None:
    """Init CLI subparsers according to registered services."""
    for command, (service, options, _) in COMMAND_SERVICE.items():
        subparser = subparsers.add_parser(
            command,
            help='Query the "{}" web-service.'.format(service),
        )
        subparser.add_argument(
            "--login",
            help="User login. Default to ENEDIS_LOGIN environment variable.",
            **arg_from_env("ENEDIS_LOGIN"),
        )
        for option_name, option_kwargs in options.items():
            subparser.add_argument(option_name, **option_kwargs)

        subparser.add_argument(
            "--cert-file",
            help="Client certificate file. Default to ENEDIS_CERT_FILE environment variable.",
            **arg_from_env("ENEDIS_CERT_FILE"),
        )
        subparser.add_argument(
            "--key-file",
            help="Client private key file. Default to ENEDIS_KEY_FILE environment variable.",
            **arg_from_env("ENEDIS_KEY_FILE"),
        )
        subparser.add_argument(
            "--homologation",
            default="ENEDIS_HOMOLOGATION" in os.environ,
            action="store_true",
            help=(
                "Change service host to use homologation sandbox. "
                "Enabled by presence of ENEDIS_HOMOLOGATION in environment variables."
            ),
        )
        subparser.add_argument(
            "--output",
            choices=["xml", "json"],
            default="xml",
            help="Output type. Default is xml",
        )


def json_encode_default(obj: Any) -> Any:
    if isinstance(obj, suds.sudsobject.Object):
        return suds.sudsobject.asdict(obj)
    elif isinstance(obj, (datetime.date, datetime.datetime)):
        return obj.isoformat()
    raise TypeError(f"Object {obj!r} is not JSON serializable")


class WSException(Exception):
    """Exception raised for ws excpetion with fault details.

    Attributes:
        exc -- the original WebFault object
        code -- the error code (if parsable)
        message -- the error message
    """

    def __init__(self, exc: WebFault):
        self.web_fault = exc
        fault = exc.fault
        try:
            res = fault.detail.erreur.resultat
            self.message = str(res.value)
            self.code: Optional[str] = str(res._code)
        except AttributeError:
            if "faultcode" in fault and "faultstring" in fault:
                self.code = str(fault.faultcode)
                self.message = str(fault.faultstring)
            else:
                # Unable to find message in 'Fault' object,
                # at least when server's response isn't properly parseable.
                self.code = None
                self.message = str(exc)

    def __str__(self) -> str:
        if self.code:
            return f"{self.code}: {self.message}"
        return self.message


def pretty_xml(input_xml: str) -> str:
    output_xml = xml.dom.minidom.parseString(input_xml).toprettyxml(indent="  ")
    return "\n".join(line for line in output_xml.split("\n") if line.strip())


def handle_cli_command(command: str, args: argparse.Namespace) -> None:
    """Run `command` service using configuration in `args`."""
    service, _, handler = COMMAND_SERVICE[command]
    client = get_client(service, args.cert_file, args.key_file, args.homologation)
    if args.output == "xml":
        client.options.retxml = True
    try:
        obj = handler(client, args)
    except WSException as exc:
        if args.output == "json":
            rich.print_json(
                json.dumps(
                    {"errcode": exc.code, "errmsg": exc.message},
                    indent=2,
                    default=json_encode_default,
                    sort_keys=True,
                )
            )
        else:
            rich.print(pretty_xml(exc.web_fault.document.str()))
        raise
    else:
        if args.output == "json":
            rich.print_json(
                json.dumps(obj, indent=2, default=json_encode_default, sort_keys=True)
            )
        else:
            rich.print(pretty_xml(obj.decode()))


def get_client(
    service: str, cert_file: str, key_file: str, homologation: bool = False
) -> Client:
    # Need custom plugin to handle `xs:choice` potentially expected in service
    # arguments. Non-handling this seems to be an outstanding suds bug, see
    # https://stackoverflow.com/questions/5963404/suds-and-choice-tag
    client = Client(
        wsdl(service),
        plugins=[_SetChoicePlugin()],
        transport=HTTPSClientCertTransport(cert_file, key_file),
    )
    # XXX unclear why this has to be done to properly consider port's
    # location. If not, options'location is set to WSDL url and override the
    # port's location.
    client.set_options(location=None)

    for method in iter_methods(client):
        # some services are still using .erdf.fr domain, but it's
        # misconfigured and use .enedis.fr certificate causing
        # verification failure.
        method.location = method.location.replace(b".erdf.fr", b".enedis.fr")

        # ConsultationMesuresDetaillees location is also misconfigured
        if (
            method.location
            == b"http://www.enedis.fr/sge/b2b/services/consultationmesuresdetaillees/v2.0"
        ):
            method.location = (
                b"https://sge-b2b.enedis.fr/ConsultationMesuresDetaillees/v2.0"
            )

        if homologation:
            method.location = method.location.replace(
                b"/sge-b2b.",
                b"/sge-homologation-b2b.",
            )

    return client


class _SetChoicePlugin(DocumentPlugin):  # type: ignore[misc]
    def __init__(self) -> None:
        self.choosen_tags = {"autorisationClient"}

    def parsed(self, context: suds.plugin.DocumentContext) -> None:
        self._rec_set_choices(context.document)

    def _rec_set_choices(self, document: suds.sax.element.Element) -> None:
        for i in document.children:
            if i.name == "choice":
                for j in i.children:
                    if j.getAttribute("name").value in self.choosen_tags:
                        i.parent.append(j)

            else:
                # recursion
                self._rec_set_choices(i)


def register(
    command: str, service: str, options: dict[str, Any]
) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """Decorator registering function as a service handler for CLI `command`,
    matching `service` web-service name, accepting `options` (list of dict given
    to `argparse.ArgumentParser.add_argument`)

    """

    def decorator(func: Callable[..., RT]) -> Callable[..., RT]:
        COMMAND_SERVICE[command] = (service, options, func)
        return func

    return decorator


def ws(service: str) -> Callable[[Callable[..., RT]], Callable[..., RT]]:
    """Decorator around sge web service call, returning a wrapper that properly set
    SOAP headers required by the service

    """

    def decorator(func: Callable[..., RT]) -> Callable[..., RT]:
        @wraps(func)
        def call_service(client: Client, args: argparse.Namespace) -> RT:
            # Headers are not consistent among webservices. If the
            # following gets more convoluted, maybe we should not try to
            # make it generic.

            # Expect a single service with a single port with a single method
            methods = client.wsdl.services[0].ports[0].methods
            assert len(methods) == 1
            method = next(iter(methods.values()))
            # Check if the method uses a header
            headers = method.soap.input.headers
            if len(headers) == 1:
                header_name, header_ns = headers[0].part.element
                header_element = "{" + header_ns + "}" + header_name
                header = client.factory.create(header_element)
                if "version" in header:
                    header.version = service.split("-")[1][1:]
                header.infoDemandeur.loginDemandeur = get_option(args, "login")
                client.set_options(soapheaders=header)

            try:
                return func(client, args)
            except WebFault as exc:
                raise WSException(exc)

        return call_service

    return decorator


def create_from_options(
    client: Client,
    args: argparse.Namespace,
    xstype_name: str,
    options_map: dict[str, str],
) -> Optional[suds.sudsobject.Object]:
    """Create and return an `xstype_name` element, and fill it according to
    `options_map` {arg name: xs element name}` mapping by looking for value in
    command line `args`.

    """
    if not hasattr(client, "xstypes_map"):
        client.xstypes_map = build_xstypes_map(client)
    xstype, prefix = client.xstypes_map[xstype_name]
    children_map = xstype_children_map(xstype)
    instance = client.factory.create("{}:{}".format(prefix, xstype_name))

    has_value = False
    for option in options_map:
        value = get_option(args, option)

        element_name = options_map[option]
        try:
            element = children_map[element_name]
        except KeyError as exc:
            exc.args = ("No element {} in {}".format(exc, list(children_map)),)
            raise

        if value is not None:
            xs_boolean = ("boolean", "http://www.w3.org/2001/XMLSchema")
            has_boolean_restriction = False
            typ = client.wsdl.schema.types.get(element.type)
            if isinstance(typ, suds.xsd.sxbasic.Simple):
                for child in typ.rawchildren:
                    if (
                        isinstance(child, suds.xsd.sxbasic.Restriction)
                        and child.ref == xs_boolean
                    ):
                        has_boolean_restriction = True
                        break
            if element.type == xs_boolean or has_boolean_restriction:
                value = "true" if value else "false"
            setattr(instance, element_name, value)
            has_value = True
        elif element.min == "0":
            if hasattr(instance, element_name):
                # factory created some default subelement (e.g. xsd:enum) but
                # enedis web-service doesn't like those empty tags
                delattr(instance, element_name)
        else:
            raise ValueError("Expecting a value for {}".format(option))

    if has_value:
        return instance
    return None


def get_option(args: Union[dict[str, Any], argparse.Namespace], option: str) -> Any:
    if isinstance(args, dict):
        return args.get(option, None)
    else:
        return getattr(args, option)


def build_xstypes_map(
    client: Client,
) -> dict[str, tuple[suds.xsd.sxbase.SchemaObject, str]]:
    """Return a dictionary of type names defined in the XML schema (without
    namespace nor prefix), associated to 2-uple `(type object, prefix)`.

    """
    assert len(client.sd) == 1, "More than one service: {}".format(client.sd)

    xstypes_map = {}
    service_def = client.sd[0]
    for xstype in service_def.types:
        assert xstype[0] is xstype[1]  # duh?!
        xstype = xstype[0]
        ns = xstype.resolve().namespace()[1]
        prefix = service_def.getprefix(ns)
        # Fix for issue #34
        if service_def.service.name != "CommanderAccesDonneesMesures-V1.0":
            assert xstype.name not in xstypes_map
        xstypes_map[xstype.name] = (xstype, prefix)

    return xstypes_map


def xstype_children_map(
    xstype: suds.xsd.sxbase.SchemaObject,
) -> dict[str, suds.xsd.sxbasic.Element]:
    """Return a dictionary of children tags for xstype, without namespace nor
    prefix.

    """
    children_map = {}
    for element, _ in xstype.children():
        children_map[element.name] = element

    return children_map


def iter_methods(client: Client) -> Iterator[suds.sudsobject.Facade]:
    """Return an iterator on exposerd methods suds instances."""
    for service in client.wsdl.services:
        for port in service.ports:
            for method in port.methods.values():
                yield method


def dict_from_dicts(*dicts: dict[str, Any]) -> dict[str, Any]:
    """Build a dictionary from multiple dictionaries. In case of key conflict, the
    latest one wins.

    """
    result = {}
    for d in dicts:
        result.update(d)

    return result



import binascii
from typing import IO


def decrypt_aes_128_cbc(key: str, iv: str, fd: IO[bytes]) -> bytes:
    """Decrypt AES 128 CBC enedis encrypted file.

    Should be equivalent to:
    openssl enc -d -aes-128-cbc -K <hex key> -iv <hex iv> -in ERDF_<..>.zip -out output.zip
    """
    import cryptography.hazmat.primitives.ciphers as crypto

    cipher = crypto.Cipher(
        crypto.algorithms.AES128(binascii.unhexlify(key)),
        crypto.modes.CBC(binascii.unhexlify(iv)),
    )
    decryptor = cipher.decryptor()
    return decryptor.update(fd.read()) + decryptor.finalize()



# Copyright (c) 2021 by Lowatt - info@lowatt.fr
#
# This program is part of lowatt_enedis
#
# lowatt_enedis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lowatt_enedis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lowatt_enedis.  If not, see <https://www.gnu.org/licenses/>.
import argparse
import sys
from typing import NoReturn

from suds.client import Client

import lowatt_enedis.services  # noqa: register services
from lowatt_enedis import (
    COMMAND_SERVICE,
    WSException,
    arg_from_env,
    handle_cli_command,
    init_cli,
    wsdl,
)

from . import decrypt


def _cli_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser(
        description="CLI access to Enedis SGE web-services.",
    )

    subparsers = parser.add_subparsers(dest="command")
    init_cli(subparsers)

    # wsdl command
    subparser = subparsers.add_parser(
        "wsdl",
        help="Display WSDL information about some ws command.",
    )
    subparser.add_argument(
        "service_command",
        help="Command name.",
    )

    # decrypt command
    subparser = subparsers.add_parser(
        "decrypt",
        help="Decrypt AES-128-CBC enedis files",
    )
    subparser.add_argument(
        "--key",
        help="Hex AES key",
        **arg_from_env("ENEDIS_DECRYPT_KEY"),
    )
    subparser.add_argument(
        "--iv",
        help="Hex CBC IV",
        **arg_from_env("ENEDIS_DECRYPT_IV"),
    )
    subparser.add_argument(
        "input_file", help="Input file", type=argparse.FileType("rb")
    )
    subparser.add_argument(
        "-o",
        "--output",
        help="Output file",
        required=True,
    )

    return parser


def run() -> NoReturn:
    parser = _cli_parser()
    args = parser.parse_args()

    if args.command == "wsdl":
        try:
            service_name = COMMAND_SERVICE[args.service_command][0]
        except KeyError:
            service_name = args.service_command

        print(Client(wsdl(service_name)))  # noqa

    elif args.command == "decrypt":
        decrypted = decrypt.decrypt_aes_128_cbc(args.key, args.iv, args.input_file)
        with open(args.output, "wb") as f:
            f.write(decrypted)
    elif args.command:
        try:
            handle_cli_command(args.command, args)
        except WSException:
            sys.exit(1)

    else:
        parser.print_help()
        sys.exit(1)

    sys.exit(0)


if __name__ == "__main__":
    run()



import binascii
import os
import sys
from pathlib import Path
from unittest import mock

import cryptography.hazmat.primitives.ciphers as crypto
import pytest

from lowatt_enedis.__main__ import run


def test_decrypt(tmp_path: Path, capsys: pytest.CaptureFixture[str]) -> None:
    key = binascii.hexlify(os.urandom(16))
    iv = binascii.hexlify(os.urandom(16))
    cipher = crypto.Cipher(
        crypto.algorithms.AES128(binascii.unhexlify(key)),
        crypto.modes.CBC(binascii.unhexlify(iv)),
    )
    encryptor = cipher.encryptor()
    encrypted = encryptor.update(b"some xml secrets") + encryptor.finalize()
    with (tmp_path / "encrypted").open("wb") as f:
        f.write(encrypted)

    with mock.patch.object(
        sys,
        "argv",
        [
            "lowatt-enedis",
            "decrypt",
            "--key",
            key.decode("ascii"),
            "--iv",
            iv.decode("ascii"),
            "--output",
            str(tmp_path / "decrypted"),
            str(tmp_path / "encrypted"),
        ],
    ), pytest.raises(SystemExit) as cm:
        run()
    assert cm.value.code == 0
    output = capsys.readouterr()
    assert output.out == ""
    assert output.err == ""
    with (tmp_path / "decrypted").open("rb") as f:
        assert f.read() == b"some xml secrets"



from typing import Any

import pytest


def pytest_addoption(parser: Any) -> None:
    parser.addoption(
        "--regen-test-data",
        action="store_true",
        default=False,
        help="Re-generate test data from actual results",
    )


@pytest.fixture
def regen_test_data(request: Any) -> bool:
    value = request.config.getoption("--regen-test-data")
    assert isinstance(value, bool)
    return value



#!/usr/bin/env python3
# coding: utf-8
#
# Copyright (c) 2019 by Lowatt info@lowatt.fr
#
# This program is part of lowatt_enedis, which is *NOT* free software.

from setuptools import setup


def local_scheme(version: str) -> str:
    """Generate a PEP440 compatible version if PEP440_VERSION is enabled"""
    import os

    import setuptools_scm.version  # only present during setup time

    if "PEP440_VERSION" in os.environ:
        return ""
    version = setuptools_scm.version.get_local_node_and_date(version)
    assert isinstance(version, str), version
    return version


if __name__ == "__main__":
    setup(use_scm_version={"local_scheme": local_scheme})



import argparse
import contextlib
import datetime
import io
import os
import sys
from typing import Iterator
from unittest import mock

import pkg_resources
import pytest
import suds.sudsobject
from suds import WebFault
from suds.client import Client, SoapClient

import lowatt_enedis as le
import lowatt_enedis.services  # noqa: register services
from lowatt_enedis.__main__ import run


def args() -> argparse.Namespace:
    return argparse.Namespace(
        cert_file=None, key_file=None, login="bob", homologation=False, output="json"
    )


@contextlib.contextmanager
def override_sys_argv(argv: list[str]) -> Iterator[None]:
    old, sys.argv = sys.argv, argv
    try:
        yield
    finally:
        sys.argv = old


def test_introspection() -> None:
    service = "RecherchePoint-v2.0"
    service_wsdl = le.wsdl(service)
    assert service_wsdl.endswith(".wsdl")

    client = Client(service_wsdl)
    xstype_map = le.build_xstypes_map(client)
    assert "CriteresType" in xstype_map

    xstype, prefix = xstype_map["CriteresType"]
    assert prefix == "ns3"

    children_map = le.xstype_children_map(xstype)
    assert "numSiret" in children_map
    assert children_map["numSiret"].min == "0"


def test_ws_decorator() -> None:
    service, options, handler = le.COMMAND_SERVICE["search"]
    service_location = None
    handler_called = False

    @le.register("testws", service, options)
    @le.ws(service)
    def test_handler(
        client: Client, args: argparse.Namespace
    ) -> suds.sudsobject.Object:
        nonlocal handler_called, service_location

        handler_called = True

        assert client.options.location is None

        headers = client.options.soapheaders
        assert headers.version == "2.0"
        assert headers.infoDemandeur.loginDemandeur == "bob"

        for service in client.wsdl.services:
            for port in service.ports:
                for method in port.methods.values():
                    assert service_location is None
                    service_location = method.location
        return suds.sudsobject.Object()

    le.handle_cli_command("testws", args())
    assert handler_called
    assert service_location == (b"https://sge-b2b.enedis.fr/RecherchePoint/v2.0")

    handler_called = False
    service_location = None
    # XXX args(homologation=True) didn't work as expected
    _args = args()
    _args.homologation = True
    le.handle_cli_command("testws", _args)
    assert handler_called
    assert service_location == (
        b"https://sge-homologation-b2b.enedis.fr/RecherchePoint/v2.0"
    )


def test_cli_output(capsys: pytest.CaptureFixture[str]) -> None:
    @le.register(
        "test",
        "RecherchePoint-v2.0",
        {
            "--do-raise": {"action": "store_true", "help": "raise a webfault"},
            "--do-raise-no-value": {
                "action": "store_true",
                "help": "raise a webfault with no value",
            },
        },
    )
    @le.ws("Test-v1.0")
    def handler(client: Client, args: argparse.Namespace) -> suds.sudsobject.Object:
        if args.do_raise:
            fault = suds.sudsobject.Object()
            fault.detail = suds.sudsobject.Object()
            fault.detail.erreur = suds.sudsobject.Object()
            fault.detail.erreur.resultat = suds.sudsobject.Object()
            fault.detail.erreur.resultat._code = "STG42"
            fault.detail.erreur.resultat.value = "some error"
            obj = suds.sax.document.Document(suds.sax.element.Element("error"))
            raise WebFault(fault=fault, document=obj)
        if args.do_raise_no_value:
            # Error as returned by M023 on homologation 23.4
            fault = suds.sudsobject.Object()
            fault.faultcode = "soap:Server"
            fault.faultstring = "Erreur Technique MFI"
            fault.detail = suds.sudsobject.Object()
            fault.detail.erreur = suds.sudsobject.Object()
            fault.detail.erreur.resultat = suds.sudsobject.Object()
            fault.detail.erreur.resultat._code = ""
            obj = suds.sudsobject.Object()
            raise WebFault(fault=fault, document=obj)
        if args.output == "xml":
            # mimic client.options.retxml = True
            return b"<data>foo</data>"
        assert args.output == "json", args
        obj = suds.sudsobject.Object()
        obj.data = "foo"
        return obj

    with mock.patch.dict(
        os.environ,
        {
            "ENEDIS_CERT_FILE": "tls.crt",
            "ENEDIS_KEY_FILE": "tls.key",
            "ENEDIS_LOGIN": "bob",
            "ENEDIS_CONTRAT": "1234",
        },
        clear=True,
    ):
        for cmd, expected_code, expected_out in (
            (
                ["lowatt-enedis", "test"],
                0,
                '<?xml version="1.0" ?>\n<data>foo</data>\n',
            ),
            (
                ["lowatt-enedis", "test", "--do-raise"],
                1,
                '<?xml version="1.0" ?>\n<error/>\n',
            ),
            (["lowatt-enedis", "test", "--output=json"], 0, '{\n  "data": "foo"\n}\n'),
            (
                ["lowatt-enedis", "test", "--do-raise", "--output=json"],
                1,
                '{\n  "errcode": "STG42",\n  "errmsg": "some error"\n}\n',
            ),
            (
                ["lowatt-enedis", "test", "--do-raise-no-value", "--output=json"],
                1,
                '{\n  "errcode": "soap:Server",\n  "errmsg": "Erreur Technique MFI"\n}\n',
            ),
        ):
            with mock.patch.object(sys, "argv", cmd), pytest.raises(SystemExit) as cm:
                run()
            assert cm.value.code == expected_code
            output = capsys.readouterr()
            assert output.out == expected_out
            assert output.err == ""


def test_cli_help() -> None:
    entrypoint = pkg_resources.get_entry_info(
        "lowatt_enedis",
        "console_scripts",
        "lowatt-enedis",
    )
    assert entrypoint is not None
    func = entrypoint.load()
    stdout = io.StringIO()
    with pytest.raises(SystemExit) as cm, override_sys_argv(
        ["lowatt-enedis", "--help"],
    ), contextlib.redirect_stdout(stdout):
        func()
    assert cm.value.code == 0
    output = stdout.getvalue()
    assert output.startswith("usage: ")


DATA_DIR = os.path.join(os.path.dirname(__file__), "data")


def test_detailed_measures_resp2py() -> None:
    service, options, handler = le.COMMAND_SERVICE["details"]
    client = Client(le.wsdl(service))
    soap = SoapClient(client, client.service.consulterMesuresDetaillees.method)
    resp_file = os.path.join(DATA_DIR, "consulterMesuresDetailleesResponse.xml")
    with open(resp_file) as stream:
        resp = soap.succeeded(soap.method.binding.input, stream.read())
        data = list(le.services.detailed_measures_resp2py(resp))
        assert len(data) == 8
        assert data[0] == (
            datetime.datetime(2020, 2, 29, 23, tzinfo=le.services.UTC),
            100,
        )


def test_measures_resp2py() -> None:
    service, options, handler = le.COMMAND_SERVICE["measures"]
    client = Client(le.wsdl(service))
    soap = SoapClient(client, client.service.consulterMesures.method)
    resp_file = os.path.join(DATA_DIR, "consulterMesuresResponse.xml")
    with open(resp_file) as stream:
        resp = soap.succeeded(soap.method.binding.input, stream.read())
        data = list(le.services.measures_resp2py(resp))
        assert data == [
            {
                "calendrier": "DI000001",
                "classeTemporelle": "BASE",
                "grandeurPhysique": "EA",
                "grille": "turpe",
                "mesures": [
                    (
                        datetime.date(2021, 2, 3),
                        datetime.date(2021, 3, 5),
                        483,
                        "REEL",
                        "CYCLIQUE",
                        "INITIALE",
                    ),
                    (
                        datetime.date(2021, 1, 5),
                        datetime.date(2021, 2, 3),
                        494,
                        "REEL",
                        "CYCLIQUE",
                        "INITIALE",
                    ),
                ],
                "unit": "kWh",
            },
            {
                "calendrier": "FC000013",
                "classeTemporelle": "BASE",
                "grandeurPhysique": "EA",
                "grille": "frn",
                "mesures": [
                    (
                        datetime.date(2021, 2, 3),
                        datetime.date(2021, 3, 5),
                        483,
                        "REEL",
                        "CYCLIQUE",
                        "INITIALE",
                    ),
                    (
                        datetime.date(2021, 1, 5),
                        datetime.date(2021, 2, 3),
                        494,
                        "REEL",
                        "CYCLIQUE",
                        "INITIALE",
                    ),
                ],
                "unit": "kWh",
            },
        ]


def test_create_from_options_boolean() -> None:
    """Test that create_from_options() outputs valid boolean values.

    suds does not seem to generate/check those values correctly,
    the only valid values for xs:boolean are "true" and "false".
    """

    # From ConsultationMesuresDetaillees-v3.0
    # <xs:complexType name="Demande">
    #     <xs:sequence>
    #         <!-- … -->
    #         <xs:element name="mesuresCorrigees" type="xs:boolean"/>
    #         <!-- … -->
    #     </xs:sequence>
    # </xs:complexType>

    service, _, _ = le.COMMAND_SERVICE["detailsV3"]
    client = Client(le.wsdl(service))
    args = argparse.Namespace()

    for value in [True, False]:
        args.corrigee = value
        option_map = {"corrigee": "mesuresCorrigees"}
        demande = le.create_from_options(client, args, "Demande", option_map)
        assert demande is not None
        assert demande.mesuresCorrigees == ("true" if value else "false")

    # From CommanderTransmissionHistoriqueMesures
    # <xs:simpleType name="BooleenType">
    #     <xs:restriction base="xs:boolean"/>
    # </xs:simpleType>
    # <xs:complexType name="DemandeHistoriqueMesuresType">
    #     <xs:sequence>
    #         <!-- … -->
    #         <xs:element name="mesureCorrigee" type="ds:BooleenType"/>
    #         <!-- … -->
    #     </xs:sequence>
    # </xs:complexType>

    service, _, _ = le.COMMAND_SERVICE["cmdHisto"]
    client = Client(le.wsdl(service))
    args = argparse.Namespace()

    for value in [True, False]:
        args.corrigee = value
        option_map = {"corrigee": "mesureCorrigee"}
        demande = le.create_from_options(
            client, args, "DemandeHistoriqueMesuresType", option_map
        )
        assert demande is not None
        assert demande.mesureCorrigee == ("true" if value else "false")



# Copyright (c) 2021 by Lowatt - info@lowatt.fr
#
# This program is part of lowatt_enedis
#
# lowatt_enedis is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lowatt_enedis is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with lowatt_enedis.  If not, see <https://www.gnu.org/licenses/>.
"""
`lowatt_enedis.certauth`
------------------------

Provide a SUDS_ transport class to connect through an HTTPS connection using
a client SSL certificate.

.. autoclass:: HTTPSClientCertTransport

.. _SUDS: https://suds-py3.readthedocs.io/en/latest

"""

import ssl
from http.client import HTTPResponse, HTTPSConnection
from typing import Any
from urllib.request import HTTPSHandler, Request, build_opener

import certifi
from suds.transport.http import HttpTransport


class _HTTPSClientAuthHandler(HTTPSHandler):
    def __init__(self, cert_file: str, key_file: str):
        super().__init__()
        self.cert_file = cert_file
        self.key_file = key_file

    def https_open(self, req: Request) -> HTTPResponse:
        return self.do_open(self.get_connection, req)  # type: ignore[arg-type]

    def get_connection(self, host: str, timeout: int = 300) -> HTTPSConnection:
        context = ssl.SSLContext(protocol=ssl.PROTOCOL_TLS_CLIENT)
        context.load_cert_chain(self.cert_file, self.key_file)
        context.load_verify_locations(cafile=certifi.where())
        return HTTPSConnection(host, context=context)


class HTTPSClientCertTransport(HttpTransport):  # type: ignore[misc]
    """SUDS_ transport class to connect through an HTTPS connection using
    a client SSL certificate.

    :param cert_file: path to the full-chain certificate file.
    :param key_file: path to the private key file.

    Usage ::

        from suds.client import Client
        client = Client(
            service_url,
            transport=HTTPSClientCertTransport(cert_file, key_file),
        )

    .. _SUDS: https://suds-py3.readthedocs.io/en/latest
    """

    def __init__(self, cert_file: str, key_file: str, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self.url_open = build_opener(
            _HTTPSClientAuthHandler(cert_file, key_file),
        ).open

    def u2open(self, u2request: Request) -> HTTPResponse:
        """Open an return a connection.

        :param u2request: A urllib2 request.
        :return: The opened file-like urllib2 object.
        """
        return self.url_open(u2request, timeout=self.options.timeout)  # type: ignore[no-any-return]
